#!/usr/bin/perl -w
#
#  Module with common functions used by the help system.
#
#  This file is part of the SuSE help system.
#
#  Copyright (C) 2002  SuSE Linux AG, Nuernberg
#
#  Author: Cornelius Schumacher <cschum@suse.de>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#use strict;

use utf8;

package susehelp;

# Variable declarations for susehelprc
our $apachedir;
our $cgidir;
our $susehelpcgidir;
our $metadir;
our $susehelpdir;
our $susehelpsymbols;
our $default_indexdir;
our $htdigsearch;
our $htdigindexer;
our $htsearchpath;
our $htdigdb;
our $htdigconf;
our $htdigbin;
our $dbg;

if ( !do "/usr/share/susehelp/susehelprc" ) {
  print STDERR "Warning! Can't read susehelprc.\n";
}

use Exporter;
use vars qw( $VERSION @ISA @EXPORT );

use substitution;

use Encode;

@ISA = qw( Exporter );
@EXPORT = qw( checkQuery parseQuery getQuery
              readDesktopFile DocEntry traverseMetaDir
              printCgiHeader printErrorPage printPage
              registerCgiPrefix registerFilePrefix 
              saveSearchState readSearchState random 
              registerMaxResultsCombo registerMethodCombo
              setLanguage printSearchResultHeader
              createCgiUrl createCgiQuery dbg
              indexDir langCharset );
              
$VERSION = 1.00;

my $currentdepth = 0;
my %messages;
my %query;

sub random()
{
  return rand();
}

&registerSymbolFunction("RANDOM","susehelp::random");

sub dbg($)
{
  $dbg && print STDERR shift, "\n";
}

sub checkQuery
{
  if ( !$ENV{ QUERY_STRING } ) {
    if ( @ARGV == 1 ) {
      $ENV{ QUERY_STRING } = $ARGV[ 0 ];
    } else {
      printErrorPage( "Error. Arguments are missing." );
      exit 1;
    }
  }
}

##
## Extract fields from "QUERY_STRING" Environment Variable (METHOD = GET)
##
sub parseQuery
{
  my $n;
  my $v;
  my @f = split ( /&/, $ENV{"QUERY_STRING"} );
  for ( my $i=0; $i<=$#f; $i++ ) {
    ( $n, $v ) = split ( /=/, $f[$i], 2 );

#    &dbg("n:$n");
    # Security sanity checks
    $v =~ s/^\-//g;			                      # Replace '-' with '' -> options
#    &dbg(" v1:$v");
    $v =~ s/%([0-9A-Fa-f][0-9A-Fa-f])/pack("c", hex($1))/eg;  # Replace %xx characters with ASCII
#    &dbg(" v2:$v");
    $v = decode( "utf8", $v );
#    &dbg(" v3:$v");
    $v =~ s/([^\p{IsWord} \-\+\/:]\(\))//g;  # remove unwanted special characters
#    &dbg(" v4:$v");

    $query{$n} = $v;
  
#    print STDERR "QUERY: $n, $v\n";
  }
}

sub getQuery
{
  my $key = shift;
  my $default = shift;

  if ( exists $query{ $key } ) {
    return $query{ $key };
  } else {
    return $default;
  }
}

use Class::Struct qw( struct );

struct 'DocEntry' => [
  filename => '$',
  identifier => '$',
  name => '$',
  comment => '$',
  docpath => '$',
  searchmethod => '$',
  search => '$',
  searchenableddefault => '$',
  indextestfile => '$',
  indexer => '$',
  weight => '$',
  children => '@',
  path => '$',
  doctype => '$',
];

# Read desktop file and set variables accordingnly. You will call this function
# with a staement similar to:
# 
#   local( $name, $comment, $docpath, $searchmethod, $search );
#   readDesktopFile( "path/to/destop/file.desktop" );
#
# Arguments:
#  path:  complete path to desktop file
#  entry: name of desktop file (without extensions)
#  lang:  language to be shown
#
# Returns:
#  $name: Translated name of the entry.
#  $comment: Translated comment about the entry.
#  $docpath: URL to documentation.
#  $searchmethod: keyword for used searchmethod.
#  $search: URL to search query.
#
sub readDesktopFile($$$)
{
  my $path = shift;
  my $entry = shift;
  my $lang = shift;

  my ( $name, $comment, $docpath, $searchmethod, $search, $identifier,
       $indextestfile, $indexer, $weight, $doctype );

#  print STDERR "DESKTOPFILE: $path, $entry, $lang\n";

  if ( !open DESKTOP, "$path" ) {
    print STDERR "Can't open '$path'\n";
    return;
  }

  my ( $commonname, $translatedname ) = ( "", "" );
  my ( $commoncomment, $translatedcomment ) = ( "", "" );

  my $docentry = DocEntry->new();
  
  while( <DESKTOP> ) {
    if ( /^X-DOC-Identifier=(.*)/ ) { $identifier = $1; }
    if ( /^Name=(.*)/ ) { $commonname = $1; }
    elsif ( /^Name\[(.?.?)\]=(.*)/ ) {
      if ( $1 eq $lang ) { $translatedname = $2; }
    } elsif ( /^DocPath=(.*)/ ) { $docpath = $1; }
    elsif ( /^Comment=(.*)/ ) { $commoncomment = $1; }
    elsif ( /^Comment\[(.?.?)\]=(.*)/ ) {
      if ( $1 eq $lang ) { $translatedcomment = $2; }
    } elsif ( /^X-DOC-SearchMethod=(.*)/ ) { $searchmethod = $1; }
    elsif ( /^X-DOC-Search=(.*)/ ) { $search = $1; }
    elsif ( /^X-DOC-IndexTestFile=(.*)/ ) {
      $indextestfile = $1;
    } elsif ( /^X-DOC-Indexer=(.*)/ ) {
      $indexer = $1;
    } elsif ( /^X-DOC-SearchEnabledDefault=(.*)/ ) {
      if ( lc( $1 ) eq "true" || lc( $1 ) eq "yes" || $1 eq "1" ) {
        $docentry->searchenableddefault( 1 );
      }
    } elsif ( /^X-DOC-Weight=(.*)/ ) {
      $weight = $1;
    } elsif ( /^X-DOC-DocumentType=(.*)/ ) {
      $doctype = $1;
    }
  }

  if ( !$identifier ) { $identifier = $entry; }

  if ( $translatedname ) { $name = $translatedname; }
  elsif ( $commonname ) { $name = $commonname; }
  else { $name = $entry; }

  if ( $translatedcomment ) { $comment = "$translatedcomment"; }
  else { $comment = "$commoncomment"; }

#  if ( lc( $searchmethod ) eq "htdig" ) {
#    if ( !$search ) { $search = $htdigsearch; }
#    if ( !$indextestfile ) {
#      $indextestfile = "$default_indexdir/$identifier.exists"; }
#    if ( !$indexer ) { $indexer = $htdigindexer; }
#  }

  if ( $doctype ) {
    $indextestfile = "$default_indexdir/$identifier.exists";
    if ( $doctype eq "text/html" ) {
      $indexer = $htdigindexer;
    } elsif ($doctype eq "text/docbook" ) {
      $indexer = $docbookindexer;
    }
  }

#  print STDERR "READ: $name\n  $comment\n  $docpath\n  $searchmethod\n  $search\n"; 

  $docentry->filename( $path );
  $docentry->identifier( $identifier );
  $docentry->name( $name );
  $docentry->comment( $comment );
  $docentry->docpath( $docpath );
  $docentry->searchmethod( $searchmethod );
  $docentry->search( $search );
  $docentry->indextestfile( $indextestfile );
  $docentry->indexer( $indexer );
  $docentry->weight( $weight );
  $docentry->doctype( $doctype );
  
  return $docentry;
}

sub traverseMetaDir($$$$$$$$)
{
  my $dirname = shift;
  my $depth = shift;
  my $lang = shift;
  my $callback_file = shift;
  my $callback_dir = shift;
  my $callback_down = shift;
  my $callback_up = shift;
  my $includemissing = shift;

  if ( $dirname && $dirname !~ /\/$/ ) { $dirname .= "/"; }

  my @docentries = scanDir( $dirname, $depth, $lang, $includemissing );

  traverseMeta( \@docentries, $callback_file, $callback_dir, $callback_down,
                $callback_up );
}

sub traverseMeta( @$$$$ )
{
  my $docentries = shift;
  my $callback_file = shift;
  my $callback_dir = shift;
  my $callback_down = shift;
  my $callback_up = shift;
  
#  print STDERR "TRAVERSE META\n";
#  print STDERR "docentries: ";
#  for my $myentry ( @$docentries ) { print STDERR "'", $myentry->name(), "' "; }
#  print STDERR "\n";

  &$callback_down;
  
  foreach my $docentry ( @$docentries ) {
    my $children = $docentry->children();
    if ( @$children ) {
      &$callback_dir( $docentry, $docentry->path() );
      traverseMeta( $children, $callback_file, $callback_dir, $callback_down,
                    $callback_up );
    } else {
      &$callback_file( $docentry, $docentry->filename() );
    }
  }

  &$callback_up;
}

sub scanDir( $$$$ )
{
  my $dirname = shift;
  my $depth = shift;
  my $lang = shift;
  my $includemissing = shift;

  my @docentries;

  opendir(DIR, "$metadir/$dirname") || die "can't opendir $metadir/$dirname: $!";
    my @entries = readdir(DIR);
    foreach my $entry (@entries) {
      if ( $entry eq "." || $entry eq ".." ) { next; }
      my $path = "$metadir/$dirname/$entry";
      my $docentry;
      if ( -d $path ) {
#        dbg("$currentdepth DIR: $entry");
#        print "<li>DIR: $entry</li>\n";
        if ( -e "$path/.directory" ) {
          $docentry = readDesktopFile( "$path/.directory", $entry, $lang );
        }
        if ( !$docentry ) {
          $docentry = DocEntry::->new();
          $docentry->name( $entry );
        }
        $docentry->path( "$dirname/$entry" );

        push @docentries, $docentry;
        
        if ( $depth > $currentdepth ) {
          $currentdepth++;
          my @subentries = scanDir( "$dirname/$entry", $depth, $lang,
                                    $includemissing );
          $currentdepth--;
          $docentry->children( \@subentries );  
        }
      } else {
#        &dbg("$currentdepth FILE: $entry");
        if ( $entry !~ /\.desktop$/ ) { next; }
#        print STDERR "ENTRY: $entry\n";
        my $id;
        if ( $entry =~ /(.*)\.(.?.?)\.desktop$/ ) {
#          print STDERR "  LANG: $2\n";
          if ( $2 ne $lang ) { next; }
          $id = $1 . ".$lang";
        } else {
#          print STDERR "  NOLANG\n";
          $entry =~ /(.*)\.[^\.]*$/;
          $id = $1;
        }
#        print STDERR "  ID: $id\n";
        $docentry = readDesktopFile( $path, $id, $lang );
        if ( !$docentry ) {
          print STDERR "Could not parse file '$path'.\n";
          next;
        }
        my $includefile = 1;
        if ( !$includemissing ) {
          my $docPath = $docentry->docpath;
#          print STDERR "DOCPATH: $docPath\n";
          if ( $docPath ) {
            if ( $docPath =~ /^\// && !-e $docPath ) {
#              print STDERR "MISSING: $docPath\n";
              $includefile = 0;
            }
          }
        }
        if ( $includefile ) {
          push @docentries, $docentry;
        }
      }
    }
  closedir DIR;

  sub compareDocEntries
  {
    $a->weight() <=> $b->weight() || $a->identifier() cmp $b->identifier
  }

  @docentries = sort compareDocEntries @docentries;

  return @docentries;
}

sub printCgiHeader()
{
  print "Content-type: text/html; charset=utf-8\n\n";
}

sub printErrorPage($)
{
  my $errorMessage = shift;

  &printCgiHeader;

  print "<html><header><title>Error</title></header><body>\n";
  print "<h1>Error</h1>\n";

  print "$errorMessage\n";

  print "</body></html>\n";

  exit 1;
}

sub printPage($)
{
  my $out = shift;

  # Substitute symbols
  readSymbols( "$susehelpdir/susehelp.symbols" );
  my $output = substituteSymbols( $out );
  print $output;
}

sub printSearchResultHeader($$)
{
  my $nummatches = shift;
  my $maxnum = shift;

  my $number;
  
  if ( !$number ) { $number = 5; }

  if ( $nummatches == 0 ) {
    print i18n("No matches found"), "\n";
  } else {
    my $lastdisplayed = $maxnum > $nummatches ? $nummatches : $maxnum;
    my $foundmsg = i18n('FIRSTDISPLAYED - LASTDISPLAYED of MATCHES matches');
    $foundmsg =~ s/FIRSTDISPLAYED/1/;
    $foundmsg =~ s/LASTDISPLAYED/$lastdisplayed/;
    $foundmsg =~ s/MATCHES/$nummatches/e;
    print $foundmsg . "\n";
  }
}

sub registerCgiPrefix($)
{
  my $prefix = shift;

  &registerSymbol( "CGIPREFIX", $prefix );
}

sub registerFilePrefix($)
{
  my $prefix = shift;

  &registerSymbol( "FILEPREFIX", $prefix );
}

sub registerMaxResultsCombo($)
{
  my $current = shift;
  
  my $combo = "<select name=\"num\" size=\"1\">\n";
  foreach my $i ( ( 5, 10, 25, 50, 1000 ) ) {
    $combo .= "  <option";
    if ( $i == $current ) {
      $combo .= " selected=\"selected\"";
    }
    $combo .= ">$i</option>\n"
  }
  $combo .= "</select>\n";

  &registerSymbol( "MAXRESULTS_COMBO", $combo );
}

sub registerMethodCombo($)
{
  my $current = shift;
  
  my $combo = "<select name=\"method\" size=\"1\">\n";
  foreach my $i ( ( "and", "or" ) ) {
    $combo .= "  <option";
    if ( $i eq $current ) {
      $combo .= " selected=\"selected\"";
    }
    $combo .= ">$i</option>\n";
  }
  $combo .= "</select>\n";

  &registerSymbol( "METHOD_COMBO", $combo );
}

sub setLanguage($;$)
{
  my $lang = shift;
  my $pofile = shift;

  if ( !$lang ) { $lang = "en"; }

  parsePo( $lang, $pofile );

  registerSymbolFunction( "I18N", "susehelp::i18n" );
}

sub i18n($)
{
  my $text = shift;

#  print STDERR "I18N: $text\n";

  if ( exists $messages{ $text } ) {
    return $messages{ $text };
  }

  return $text;
}

sub parsePo($$)
{
  my $language = shift;
  my $filename = shift;

  if ( !$filename ) {
    $filename = "susehelp.po";
  }

  if ( !-e $filename ) {
    $filename = "$susehelpdir/i18n/$language/$filename";
  }

  $language =~ s/\.\.//g;
  $filename =~ s/\.\.//g;

  if ( !open( POFILE, "<$filename") ) {
    print STDERR "Warning! Could not open '$filename'.\n";
    return;
  }

  my $id;
  my $str;

  my $line = <POFILE>;

  while ( !eof(POFILE) ) {
    if ( $line =~ m/^msgid "(.*)"/ ) {
      $id = "";
      $id = $1 if ( $1 );
      $id .= $1 while ( ( $line = <POFILE> ) =~ m/^"(.*)"/ );
    }

    if ( $line =~ m/^msgstr "(.*)"/ ) {
      $str = "";
      $str = $1 if ( $1 );
      $str .= $1 while ( ( $line = <POFILE> ) =~ m/^"(.*)"/ );
      $messages{ $id } = $str;
    }
    else { $line = <POFILE> };
  }

  close(POFILE);
}

sub createCgiQuery($;$)
{
  my $urlquery = shift;
  my $urllang = shift;

  my @urlquery = split /&/, $urlquery;

  my %args;  
  foreach ( @urlquery ) {
#    print STDERR "QUERYPART: $_\n";
    if ( /(.*)=(.*)/ ) { $args{ $1 } = $2; }
  }

  my @args;
  if ( !$args{ "cgiprefix" } ) {
    my $cgiprefix = getQuery( "cgiprefix", "cgi:" );
    push @args, "cgiprefix=$cgiprefix";
  }
  if ( !$args{ "fileprefix" } ) {
    my $fileprefix = getQuery( "fileprefix", "file:" );
    push @args, "fileprefix=$fileprefix";
  }
  if ( !$args{ "header" } ) {
    my $header = getQuery( "header", 0 );
    push @args, "header=$header";
  }
  if ( !$args{ "depth" } ) {
    my $depth = getQuery( "depth", 1 );
    push @args, "depth=$depth";
  }
  if ( !$args{ "lang" } ) {
    if ( $urllang ) { push @args, "lang=$urllang"; }
  }

  my $newquery = join "&", @args;

  if ( $urlquery ) { $newquery .= "&$urlquery"; }
  
  return $newquery;
}

sub createCgiUrl($;$)
{
  my $urlpath = shift;
  my $urllang = shift;

  my $urlbin;
  my $urlquery;
  
  if ( $urlpath =~ /(.*)\?(.*)/ ) {  
    $urlbin = $1;
    $urlquery = $2;
  } else {
    $urlbin = $urlpath;
  }

  my $newquery = createCgiQuery( $urlquery, $urllang );

  my $cgiprefix = getQuery( "cgiprefix", "cgi:" );

  my $url = "$cgiprefix$urlbin?$newquery";
  
  return $url;
}

sub indexDir()
{
  # TODO: Parse config files containing indexDir settings.
}

# Return charset used for given language
sub langCharset( $ )
{
  my $lang = shift;
  if ( $lang eq "cz" || $lang eq "hu" ) {
    return "latin2";
  } elsif ( $lang eq "kr" ) {
    return "utf8";
  } else {
    return "latin1";
  }
}

1;
